/*----------------------------------------------------------------------------
    corrslite - a lightweight correlation visualization

    Copyright (c) 2007 Ge Wang, Jonathan Abel, Miriam Kolar
        All rights reserved.
        http://ccrma.stanford.edu/~ge/software/corrslite/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    U.S.A.
-----------------------------------------------------------------------------*/

//-----------------------------------------------------------------------------
// name: corrslite.cpp
// desc: lightweight correlation demo/visualization, partly based on 'sndpeek'
//
// authors: Ge Wang (ge@ccrma.stanford.edu)
//          Jonathan Abel (abel@ccrma.stanford.edu)
//          Miriam Kolar (kolar@ccrma.stanford.edu)
// library:
//          (STK) Perry R. Cook (prc@cs.princeton.edu)
//          (STK) Gary P. Scavone (gary@ccrma.stanford.edu)
//          (FFT) CARL CMusic Distribution
//          (libsndfile) Erik de Castro Lopo
//
// usage: (type 'h' while running or see list of command line arguments)
//-----------------------------------------------------------------------------
#include <math.h>
#include <stdio.h>

// global def
#include "corrslite_def.h"

// libsndfile
#if defined(__USE_SNDFILE_NATIVE__)
  #include <sndfile.h>
#else
  #include "util_sndfile.h"
#endif

// STK
#include "RtAudio.h"
// threads
#include "util_thread.h"
// FFT
#include "chuck_fft.h"

// OpenGL
#if defined(__PLATFORM_MACOSX__)
  #include <GLUT/glut.h>
  #include <OpenGL/gl.h>
  #include <OpenGL/glu.h>
#else
  #include <GL/glut.h>
  #include <GL/gl.h>
  #include <GL/glu.h>
#endif

// STL
#include <string>
using namespace std;




#define SAMPLE              float
//-----------------------------------------------------------------------------
// filter class
// desc: ...
//-----------------------------------------------------------------------------
struct Filter
{
public:
    Filter();
    ~Filter();
    
    void init( float *B_, int lenb_, float *A_, int lena_ );
    void clear();
    void apply( SAMPLE * buffer, long buffer_size );
    void clear_sigs(); // clears the windows/signals but retains coefficients
    
    // filter coefficients
    float *A;
    float *B;
    int lena, lenb;
    
    // signals
    SAMPLE *winxp, *ys;
    int len;
};




//-----------------------------------------------------------------------------
// global variables and #defines
//-----------------------------------------------------------------------------
// number of input channels
long g_channels_in = 2;
// number of out channels
long g_channels_out = 0;
// audio I/O buffer size
long g_buffer_size = 1024;
// instance
RtAudio * g_audio = NULL;
// buffer mutex
XMutex g_mutex;
// dc blocker
Filter g_dcbloke_left;
Filter g_dcbloke_right;
// buffers
SAMPLE * g_left = NULL;
SAMPLE * g_right = NULL;
SAMPLE * g_fbuf = NULL;
SAMPLE * g_gbuf = NULL;
SAMPLE * g_buffy = NULL;
SAMPLE * g_swap = NULL;
SAMPLE ** g_slayers = NULL;
long * g_slayer_draw = NULL;
// FFT size for xcorr
long g_fft_size = 0;
// # of slayers
long g_num_slayers = 64;
// current slayer
long g_current_slayer = 0;
// xcorr view (zoom)
GLfloat g_zoom_percent = 1;
// corr scaling
GLfloat g_scaling = 1.0f;
// time 
GLfloat g_tscaling = 1.0f;
// default sample rate
#if defined(__PLATFORM_LINUX__)
  long g_srate = 48000;
#else
  long g_srate = 44100;
#endif
// dc blocker cutoff
long g_lowcut = 300;

// width and height of the window
GLsizei g_width = 800;
GLsizei g_height = 600;
GLsizei g_last_width = g_width;
GLsizei g_last_height = g_height;
// fill mode
GLenum g_fillmode = GL_FILL;
// light 0 position
GLfloat g_light0_pos[4] = { 2.0f, 1.2f, 4.0f, 1.0f };
// light 1 parameters
GLfloat g_light1_ambient[] = { .2f, .2f, .2f, 1.0f };
GLfloat g_light1_diffuse[] = { 1.0f, 1.0f, 1.0f, 1.0f };
GLfloat g_light1_specular[] = { 1.0f, 1.0f, 1.0f, 1.0f };
GLfloat g_light1_pos[4] = { -2.0f, 0.0f, -4.0f, 1.0f };
// modelview stuff
GLfloat g_angle_y = 0.0f;
GLfloat g_inc = 0.0f;
GLfloat g_eye_y = 0.0f;
// fftshift
GLboolean g_fftshift = TRUE;
GLboolean g_fullscreen = FALSE;
GLboolean g_many_slayers = FALSE;
GLboolean g_xcorr = TRUE;
GLboolean g_time_buffers = TRUE;
GLboolean g_acorr_left = TRUE;

// rotation increments, set to defaults
GLfloat g_inc_val_mouse = 0.1f;
GLfloat g_inc_val_kb = .025f;




//-----------------------------------------------------------------------------
// function prototypes
//-----------------------------------------------------------------------------
void idleFunc( );
void displayFunc( );
void reshapeFunc( int width, int height );
void keyboardFunc( unsigned char, int, int );
void mouseFunc( int button, int state, int x, int y );
void initialize_graphics( );
bool initialize_audio( );
unsigned long next_power_2( unsigned long n );
void xcorr( SAMPLE * f, int fsize, SAMPLE * g, int gsize, SAMPLE * buffy, int size );
void xcorr_fft( SAMPLE * f, int fsize, SAMPLE * g, int gsize, SAMPLE * buffy, int size );
void normalize( SAMPLE * buffy, int size, SAMPLE scale = 1.0f );
void draw_xcorr();
void draw_buffers();




//-----------------------------------------------------------------------------
// name: help()
// desc: ...
//-----------------------------------------------------------------------------
void help()
{
    fprintf( stderr, "----------------------------------------------------\n" );
    fprintf( stderr, "corrslite (1.0)\n" );
    fprintf( stderr, "http://ccrma.stanford.edu/~ge/software/corrslite/\n" );
    fprintf( stderr, "----------------------------------------------------\n" );
    fprintf( stderr, "'h' - print this help message\n" );
    fprintf( stderr, "'t' - toggle cross/auto correlation\n" );
    fprintf( stderr, "'1' - select left channel for auto-correlation\n" );
    fprintf( stderr, "'2' - select right channel for auto-correlation\n" );
    fprintf( stderr, "'-' - scale down corr-gram\n" );
    fprintf( stderr, "'=' - scale up corr-gram\n" );
    fprintf( stderr, "'_' - scale down waveform\n" );
    fprintf( stderr, "'+' - scale up waveform\n" );
    fprintf( stderr, "'<' - zoom in\n" );
    fprintf( stderr, "'>' - zoom out\n" );
    fprintf( stderr, "'[' - rotate up\n" );
    fprintf( stderr, "']' - rotate down\n" );
    fprintf( stderr, "'F' - toggle FFT shift\n" );
    fprintf( stderr, "'w' - toggle layers view\n" );
    fprintf( stderr, "'q' - quit\n" );
    fprintf( stderr, "----------------------------------------------------\n" );
    fprintf( stderr, "\n" );
}




//-----------------------------------------------------------------------------
// name: usage()
// desc: ...
//-----------------------------------------------------------------------------
void usage()
{
    fprintf( stderr, "usage: corrslite  --[options] [filename]\n" );
    fprintf( stderr, "corrslite version: 1.0\n" );
    fprintf( stderr, "\n" );
}




//-----------------------------------------------------------------------------
// Name: main( )
// Desc: entry point
//-----------------------------------------------------------------------------
int main( int argc, char ** argv )
{
    // remember command line
    GLboolean set_play = FALSE;

    // command line arguments
    for( int i = 1; i < argc; i++ )
    {
        if( !strncmp( argv[i], "-", 1 ) )
        {
            if( !strncmp(argv[i], "--srate:", 8) )
                g_srate = atoi( argv[i]+8 ) > 0 ? atoi( argv[i]+8 ) : g_srate;
            else if( !strncmp(argv[i], "--in:", 5) )
                g_channels_in = atoi( argv[i]+5 ) > 0 ? atoi( argv[i]+5 ) : g_channels_in;
        }
        else
        {
            // check for filename
        }
    }

#ifdef __OS_MACOSX__
    //save working dir
    char * cwd = getcwd( NULL, 0 );
#endif

    // initialize GLUT
    glutInit( &argc, argv );

#ifdef __OS_MACOSX__
    //restore working dir
    chdir( cwd );
    free( cwd );
#endif

    // double buffer, use rgb color, enable depth buffer
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    // initialize the window size
    glutInitWindowSize( g_width, g_height );
    // set the window postion
    glutInitWindowPosition( 100, 100 );
    // create the window
    glutCreateWindow( "corrslite");
    // full screen
    //if( g_fullscreen )
    //    glutFullScreen();

    // set the idle function - called when idle
    glutIdleFunc( idleFunc );
    // set the display function - called when redrawing
    glutDisplayFunc( displayFunc );
    // set the reshape function - called when client area changes
    glutReshapeFunc( reshapeFunc );
    // set the keyboard function - called on keyboard events
    glutKeyboardFunc( keyboardFunc );
    // set the mouse function - called on mouse stuff
    glutMouseFunc( mouseFunc );

    // do our own initialization
    initialize_graphics( );

    // initialize real-time audio
    if( !initialize_audio( ) )
    {
        fprintf( stderr, "[jackpeek]: exiting...\n" );
        return -3;
    }
    
    // print help
    help();

    // let GLUT handle the current thread from here
    glutMainLoop();

    return 0;
}




//-----------------------------------------------------------------------------
// name: cb()
// desc: audio callback
//-----------------------------------------------------------------------------
int cb( char * buffer, int buffer_size, void * user_data )
{
    SAMPLE * s = (SAMPLE *)buffer;

    // lock
    g_mutex.acquire();

    // get stereo channels
    for( long i = 0; i < buffer_size; i++ )
    {
         g_left[i] = s[i*2];
         g_right[i] = s[i*2+1];
    }

    // high pass filter
    g_dcbloke_left.apply( g_left, buffer_size );
    g_dcbloke_right.apply( g_right, buffer_size );

    // unlock
    g_mutex.release();

    return 0;
}




//-----------------------------------------------------------------------------
// Name: initialize_audio( )
// Desc: set up audio capture and playback and initializes any application data
//-----------------------------------------------------------------------------
bool initialize_audio( )
{
    // initialize rtaudio
    try
    {
        // buffer size
        int bufsize = g_buffer_size;
        // open the audio device for capture and playback
        g_audio = new RtAudio( 0, g_channels_out, 0, g_channels_in, RTAUDIO_FLOAT32,
            g_srate, &bufsize, 8 );
        // set buffer size
        g_buffer_size = bufsize;
    }
    catch( RtError & e )
    {
        // exception
        fprintf( stderr, "[corrslite](via RtAudio): %s\n", e.getMessage().c_str() );
        fprintf( stderr, "[corrslite]: error: cannot open audio device for capture/playback...\n" );
        return false;
    }

    // set the audio callback
    g_audio->setStreamCallback( cb, NULL );

    // audio buffers
    g_left = new SAMPLE[g_buffer_size];
    g_right = new SAMPLE[g_buffer_size];
    // make buffers to hold kernel and signal
    g_fft_size = next_power_2( g_buffer_size * 2 - 1 );
    g_fbuf = new SAMPLE[g_fft_size];
    g_gbuf = new SAMPLE[g_fft_size];
    g_buffy = new SAMPLE[g_fft_size];
    g_swap = new SAMPLE[g_fft_size];
    g_slayers = new SAMPLE *[g_num_slayers];
    g_slayer_draw = new long[g_num_slayers];
    for( long i = 0; i < g_num_slayers; i++ )
    {
        g_slayers[i] = new SAMPLE[g_fft_size];
        memset( g_slayers[i], 0, sizeof(SAMPLE)*g_fft_size );
    }
    memset( g_slayer_draw, 0, sizeof(long)*g_num_slayers );

    // initialize our filter
    float a1 = (ONE_PI*g_lowcut/g_srate-1) / (ONE_PI*g_lowcut/g_srate+1);
    float g = (1-a1)/2;
    float B[] = { g, -g };
    float A[] = { 1.0f, a1 };
    //float B[] = { .9791f, -.9791f };
    //float A[] = { 1.f, -.9582f };
    g_dcbloke_left.init( B, 2, A, 2 );
    g_dcbloke_right.init( B, 2, A, 2 );

    // start the audio
    g_audio->startStream( );

    return true;
}




//-----------------------------------------------------------------------------
// Name: initialize_graphics( )
// Desc: sets initial OpenGL states and initializes any application data
//-----------------------------------------------------------------------------
void initialize_graphics()
{
    // set the GL clear color - use when the color buffer is cleared
    glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );
    // set the shading model to 'smooth'
    glShadeModel( GL_SMOOTH );
    // enable depth
    glEnable( GL_DEPTH_TEST );
    // set the front faces of polygons
    glFrontFace( GL_CCW );
    // set fill mode
    glPolygonMode( GL_FRONT_AND_BACK, g_fillmode );
    // enable lighting
    glEnable( GL_LIGHTING );
    // enable lighting for front
    glLightModeli( GL_FRONT_AND_BACK, GL_TRUE );
    // material have diffuse and ambient lighting 
    glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
    // enable color
    glEnable( GL_COLOR_MATERIAL );
    // vertex
    glEnable( GL_NORMALIZE );

    // enable light 0
    glEnable( GL_LIGHT0 );

    // setup and enable light 1
    glLightfv( GL_LIGHT1, GL_AMBIENT, g_light1_ambient );
    glLightfv( GL_LIGHT1, GL_DIFFUSE, g_light1_diffuse );
    glLightfv( GL_LIGHT1, GL_SPECULAR, g_light1_specular );
    glEnable( GL_LIGHT1 );
}




//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( int w, int h )
{
    // save the new window size
    g_width = w; g_height = h;
    // map the view port to the client area
    glViewport( 0, 0, w, h );
    // set the matrix mode to project
    glMatrixMode( GL_PROJECTION );
    // load the identity matrix
    glLoadIdentity( );
    // create the viewing frustum
    gluPerspective( 45.0, (GLfloat) w / (GLfloat) h, 1.0, 300.0 );
    // set the matrix mode to modelview
    glMatrixMode( GL_MODELVIEW );
    // load the identity matrix
    glLoadIdentity( );
    // position the view point
    gluLookAt( 0.0f, 3.5f * sin( g_eye_y ), 3.5f * cos( g_eye_y ), 
               0.0f, 0.0f, 0.0f, 
               0.0f, ( cos( g_eye_y ) < 0 ? -1.0f : 1.0f ), 0.0f );

    // set the position of the lights
    glLightfv( GL_LIGHT0, GL_POSITION, g_light0_pos );
    glLightfv( GL_LIGHT1, GL_POSITION, g_light1_pos );
}




//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
    switch( key )
    {
    case 'q':
        exit( 0 );
    break;
    case 'h':
        help();
    break;
    case 's':
    {
        if( !g_fullscreen )
        {
            g_last_width = g_width;
            g_last_height = g_height;
            glutFullScreen();
        }
        else
            glutReshapeWindow( g_last_width, g_last_height );

        g_fullscreen = !g_fullscreen;
        fprintf( stderr, "[corrslite]: fullscreen:%s\n", g_fullscreen ? "ON" : "OFF" );
    }
    break;
    case 'w':
    {
        g_many_slayers = !g_many_slayers;
        fprintf( stderr, "[corrslite]: many slayers:%s\n", g_many_slayers ? "ON" : "OFF" );
    }
    break;
    case 't':
    {
        g_xcorr = !g_xcorr;
        fprintf( stderr, "[corrslite]: correlation:%s\n", g_xcorr ? "CROSS" : "AUTO" );
    }
    break;
    case '1':
    {
        g_acorr_left = true;
        fprintf( stderr, "[corrslite]: auto-corr channel:%s\n", g_acorr_left ? "1" : "2" );
    }
    break;
    case '2':
    {
        g_acorr_left = false;
        fprintf( stderr, "[corrslite]: auto-corr channel:%s\n", g_acorr_left ? "1" : "2" );
    }
    break;
    case '-': // scale down
        g_scaling /= 1.01f;
        fprintf( stderr, "[corrslite]: corr-gram scaling:%f\n", g_scaling );
    break;
    case '=': // scale up
        g_scaling *= 1.01f;
        fprintf( stderr, "[corrslite]: corr-gram scaling:%f\n", g_scaling );
    break;
    case '_': // scale down
        g_tscaling /= 1.01f;
        fprintf( stderr, "[corrslite]: waveform scaling:%f\n", g_tscaling );
    break;
    case '+': // scale up
        g_tscaling *= 1.01f;
        fprintf( stderr, "[corrslite]: waveform scaling:%f\n", g_tscaling );
    break;
    case '<': // zoom in
        g_zoom_percent /= 1.03f;
        fprintf( stderr, "[corrslite]: zoom percent:%f\n", g_zoom_percent );
    break;
    case '>': // zoom out
        g_zoom_percent *= 1.03f;
        if( g_zoom_percent > 1.0f ) g_zoom_percent = 1.0f;
        fprintf( stderr, "[corrslite]: zoom percent:%f\n", g_zoom_percent );
    break;
    case '[':
        g_eye_y -= g_inc_val_kb;
        fprintf( stderr, "[corrslite]: yview:%f\n", g_eye_y );
    break;
    case ']':
        g_eye_y += g_inc_val_kb;
        fprintf( stderr, "[corrslite]: yview:%f\n", g_eye_y );
    break;
    case 'F': // fftshift
        g_fftshift = !g_fftshift;
        fprintf( stderr, "[corrslite]: fftshift:%s\n", g_fftshift ? "ON" : "OFF" );
    break;
    }

    // do a reshape since g_eye_y might have changed
    reshapeFunc( g_width, g_height );
    glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: mouseFunc( )
// Desc: handles mouse stuff
//-----------------------------------------------------------------------------
void mouseFunc( int button, int state, int x, int y )
{
    if( button == GLUT_LEFT_BUTTON )
    {
        // rotate
        if( state == GLUT_DOWN )
            g_inc -= g_inc_val_mouse;
        else
            g_inc += g_inc_val_mouse;
    }
    else if ( button == GLUT_RIGHT_BUTTON )
    {
        if( state == GLUT_DOWN )
            g_inc += g_inc_val_mouse;
        else
            g_inc -= g_inc_val_mouse;
    }
    else
        g_inc = 0.0f;

    glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
    // render the scene
    glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// name: draw_string()
// desc: ...
//-----------------------------------------------------------------------------
void draw_string( GLfloat x, GLfloat y, GLfloat z, const char * str, GLfloat scale = 1.0f )
{
    GLint len = strlen( str ), i;

    glPushMatrix();
    glTranslatef( x, y, z );
    glScalef( .001f * scale, .001f * scale, .001f * scale );
    for( i = 0; i < len; i++ )
        glutStrokeCharacter( GLUT_STROKE_ROMAN, str[i] );
    glPopMatrix();
}




//-----------------------------------------------------------------------------
// name: next_power_2()
// desc: ...
// thanks: to Niklas Werner / music-dsp
//-----------------------------------------------------------------------------
unsigned long next_power_2( unsigned long n )
{
    unsigned long nn = n;
    for( ; n &= n-1; nn = n );
    return nn * 2;
}




//-----------------------------------------------------------------------------
// name: xcorr_fft()
// desc: FFT-based cross correlation
//-----------------------------------------------------------------------------
void xcorr_fft( SAMPLE * f, int fsize, SAMPLE * g, int gsize, SAMPLE * buffy, int size )
{
    // sanity check
    assert( (fsize + gsize - 1) == size );

    // clear
    memset( g_fbuf, 0, sizeof(SAMPLE) * g_fft_size );
    memset( g_gbuf, 0, sizeof(SAMPLE) * g_fft_size );

    // copy in
    memcpy( g_fbuf, f, sizeof(SAMPLE) * fsize );
    memcpy( g_gbuf, g, sizeof(SAMPLE) * gsize );

    // take fft
    rfft( g_fbuf, g_fft_size/2, FFT_FORWARD );
    rfft( g_gbuf, g_fft_size/2, FFT_FORWARD );

    // complex
    complex * F = (complex *)g_fbuf;
    complex * G = (complex *)g_gbuf;
    complex * Y = (complex *)buffy;

    // loop
    for( int i = 0; i < g_fft_size/2; i++ )
    {
        // conjugate F
        F[i].im = -F[i].im;
        // complex multiply
        Y[i].re = F[i].re*G[i].re - F[i].im*G[i].im;
        Y[i].im = F[i].im*G[i].re + F[i].re*G[i].im;
    }

    // inverse fft
    rfft( buffy, g_fft_size/2, FFT_INVERSE );
}




//-----------------------------------------------------------------------------
// name: xcorr()
// desc: cross correlation
//-----------------------------------------------------------------------------
void xcorr( SAMPLE * f, int fsize, SAMPLE * g, int gsize, SAMPLE * buffy, int size )
{
    // sanity check
    assert( (fsize + gsize - 1) == size );

    // clear
    memset( g_fbuf, 0, sizeof(SAMPLE) * g_fft_size );
    memset( g_gbuf, 0, sizeof(SAMPLE) * g_fft_size );

    // copy in
    memcpy( g_fbuf, f, sizeof(SAMPLE) * fsize );
    memcpy( g_gbuf, g, sizeof(SAMPLE) * gsize );

    long i, j;
    for( i = 0; i < size; i++ )
    {
        buffy[i] = 0;
        for( j = 0; (i+j) < size; j++ )
        {
            buffy[i] += g_fbuf[j] * g_gbuf[i+j];
        }
    }
}




//-----------------------------------------------------------------------------
// name: normalize()
// desc: ...
//-----------------------------------------------------------------------------
void normalize( SAMPLE * buffy, int size, SAMPLE scale )
{
    SAMPLE max = 0;

    // loop over the signal
    for( int i = 0; i < size; i++ )
        if( fabs(buffy[i]) > max )
            max = fabs( buffy[i] );

    for( int i = 0; i < size; i++ )
        buffy[i] = (buffy[i] / max) * scale;
}




//-----------------------------------------------------------------------------
// name: normalize2()
// desc: ...
//-----------------------------------------------------------------------------
void normalize2( SAMPLE * buffy, int size, SAMPLE * f, int fs, SAMPLE * g, int gs )
{
    float sum = 0.000001f;

    // f^2(t)
    for( long i = 0; i < fs; i++ )
        sum += f[i]*f[i];
    // g^2(t)
    for( long j = 0; j < gs; j++ )
        sum += g[j]*g[j];
    // normalize: taking coherence into account
    for( long k = 0; k < size; k++ )
        buffy[k] /= sum;
}




//-----------------------------------------------------------------------------
// Name: draw_buffers( )
// Desc: draw time domain
//-----------------------------------------------------------------------------
void draw_buffers( )
{
    // set
    glColor3f( .3f, .6f, 1.0f );
    glNormal3f( 0, 0, 1.0f );

    // lock
    g_mutex.acquire();

    // draw?
    if( g_time_buffers )
    {
        GLfloat mag = 1.85f;
        GLfloat xcoord = -mag/2;
        long len = g_buffer_size;
        GLfloat inc = mag / len;

        if( g_xcorr || g_acorr_left )
        {
            glBegin( GL_LINE_STRIP );
            for( long i = 0; i < g_buffer_size; i++ )
            {
                glVertex2f( xcoord, g_left[i] * .8f * g_tscaling );
                xcoord += inc;
            }
            glEnd();
        }

        xcoord = -mag/2;
        if( g_xcorr || !g_acorr_left )
        {
            glBegin( GL_LINE_STRIP );
            for( long j = 0; j < g_buffer_size; j++ )
            {
                glVertex2f( xcoord, g_right[j] * .8f * g_tscaling - .4f );
                xcoord += inc;
            }
            glEnd();
        }
    }

    // unlock
    g_mutex.release();
}




//-----------------------------------------------------------------------------
// Name: draw_xcorr( )
// Desc: draw cross correlation
//-----------------------------------------------------------------------------
void draw_xcorr( )
{
    // lock
    g_mutex.acquire();

    if( g_xcorr )
    {
        // take corr
        xcorr_fft( g_left, g_buffer_size, g_right, g_buffer_size, 
                g_buffy, g_buffer_size*2 - 1 );
        // normalize
        normalize2( g_buffy, g_fft_size, g_left, g_buffer_size,
                    g_right, g_buffer_size );
    }
    else
    {
        SAMPLE * signal = g_acorr_left ? g_left : g_right;
        // take corr
        xcorr_fft( signal, g_buffer_size, signal, g_buffer_size, 
                g_buffy, g_buffer_size*2 - 1 );
        // normalize
        normalize2( g_buffy, g_fft_size, signal, g_buffer_size,
                    signal, g_buffer_size );
    }

    // unlock
    g_mutex.release();

    // fftshift
    if( g_fftshift )
    {
        // copy and swap
        memcpy( g_swap, g_buffy+g_buffer_size, sizeof(SAMPLE) * g_buffer_size );
        memcpy( g_swap+g_buffer_size, g_buffy, sizeof(SAMPLE) * g_buffer_size );
        memcpy( g_buffy, g_swap, sizeof(SAMPLE) * g_fft_size );
    }

    // copy into current slayer
    memcpy( g_slayers[g_current_slayer], g_buffy, sizeof(SAMPLE) * g_fft_size );
    // mark this layer to draw
    g_slayer_draw[g_current_slayer] = true;

    // save the current matrix state
    glPushMatrix();

    // color waveform
    glColor3f( 1.0f, 1.0f, .25f );
    // set vertex normals (for somewhat controlled lighting)
    glNormal3f( 0.0f, 0.0f, 1.0f );

    GLfloat mag, xcoord, zcoord, inc, zinc;
    long len, start, end, n;
    // draw layers
    for( long j = 0; j < g_num_slayers; j++ )
    {
        mag = 3.7f;
        xcoord = -mag/2;
        zcoord = 0;
        n = (g_current_slayer-j);
        if( n < 0 ) n += g_num_slayers;
        zinc = .50f;
        if( g_xcorr )
        {
            len = (long)(g_fft_size * g_zoom_percent);
            start = (g_fft_size-len)/2;
            end = start + len;
            inc = mag / len;
        }
        else
        {
            len = (long)(g_buffer_size * g_zoom_percent);
            start = g_buffer_size;
            end = start + len;
            inc = mag / len;
        }

        if( g_slayer_draw[n] == false ) continue;
        //long step = len / 1024; if( step == 0 ) step = 1;
        //if( len > 1024 ) inc *= len / 1024;

        glBegin( GL_LINE_STRIP );
        for( long i = start; i < end; i++ )
        {
            glVertex3f( xcoord, g_slayers[n][i] * 3200 * g_scaling, zcoord - (j*zinc) );
            xcoord += inc;
        }
        glEnd();
    }

    // turn off
    if( !g_many_slayers ) g_slayer_draw[g_current_slayer] = false;
    // advance
    g_current_slayer++; if( g_current_slayer >= g_num_slayers ) g_current_slayer = 0;

    // restore previous matrix state
    glPopMatrix();
}




//-----------------------------------------------------------------------------
// Name: displayFunc( )
// Desc: callback function invoked to draw the client area
//-----------------------------------------------------------------------------
void displayFunc( )
{
    // clear the color and depth buffers
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    // save state
    glPushMatrix();

    // rotate the sphere about y axis
    glRotatef( g_angle_y += g_inc, 0.0f, 1.0f, 0.0f );

    glPushMatrix();
        glTranslatef( 0.0f, 1.05f, 0.0f );
        draw_buffers();
    glPopMatrix();

    glPushMatrix();
        glTranslatef( 0.0f, -.5f, 0.0f );
        draw_xcorr();
    glPopMatrix();

    // pop state
    glPopMatrix();

    // flush gl commands
    glFlush( );
    // swap the buffers
    glutSwapBuffers( );
}




Filter::Filter()
{
    A = B = NULL;
    winxp = ys = NULL;
    lena = lenb = len = 0;
}


Filter::~Filter()
{
    this->clear();
}


void Filter::init( float *B_, int lenb_, float *A_, int lena_ )
{
    assert( lenb_ != 0 && lena_ != 0 );
    this->clear();

    lenb = lenb_;
    lena = lena_;

    B = new float[lenb];
    A = new float[lena];

    memcpy( B, B_, lenb * sizeof(float) );
    memcpy( A, A_, lena * sizeof(float) );

    // normalize by A[0]
    assert( A[0] != 0.0f );
    int j;
    if( A[0] != 1.0f )
    {
        for( j = 0; j < lenb; j++ )
            B[j] /= A[0];
        for( j = 1; j < lena; j++ )
            A[j] /= A[0];
        A[0] = 1.0f;
    }

}


void Filter::clear()
{
    if( A ) delete [] A;
    if( B ) delete [] B;
    if( winxp ) delete [] winxp;
    if( ys ) delete [] ys;

    A = B = NULL;
    winxp = ys = NULL;
    lena = lenb = len = 0;
}


void Filter::clear_sigs()
{
    if( winxp ) delete winxp; 
    winxp = NULL;
    if( ys ) delete ys;
    ys = NULL;
    len = 0;
}


void Filter::apply( SAMPLE * buffer, long buffer_size )
{
    int j;

    if( !winxp || len != buffer_size )
    {
        clear_sigs();
        len = buffer_size;
        winxp = new SAMPLE[len];
        ys = new SAMPLE[len];
        memset( winxp, 0, len * sizeof(SAMPLE) );
        memset( ys, 0, len * sizeof(SAMPLE) );
    }

    for( int i = 0; i < len; i++ )
    {
        ys[i] = 0.0f;
        for( j = 0; j < lenb; j++ )
        {
            if( i-j >= 0 )
                ys[i] += B[j] * buffer[i-j];
                // y[i] += B[j] * ( (i-j) >= 0 ? x[i-j] : 0.0f );
            else
                ys[i] += B[j] * winxp[len+i-j];
        }

        for( j = 1; j < lena; j++ )
        {
            if( i-j >= 0 )
                ys[i] -= A[j] * ys[i-j];
            else
                ys[i] -= A[j] * ys[len+i-j];
        }
    }

    // copy input into winxp for next frame
    memcpy( winxp, buffer, len * sizeof(SAMPLE) );
    // copy output into the frame
    memcpy( buffer, ys, len * sizeof(SAMPLE) );
}
